import responses
from brian_dashboard_manager.grafana import dashboard, provision
from brian_dashboard_manager.grafana.utils.request import TokenRequest


@responses.activate
def test_get_dashboard(mock_grafana):
    mock_grafana.create_dashboard({"uid": "1"})

    request = mock_grafana.request

    data = dashboard._get_dashboard(request, "1")
    assert data["uid"] == "1"

    data = dashboard._get_dashboard(request, "2")
    assert data is None


@responses.activate
def test_delete_dashboards_new(mock_grafana):
    mock_grafana.create_dashboard({"uid": "1"})
    mock_grafana.create_dashboard({"uid": "2"})

    request = mock_grafana.request
    assert dashboard.delete_dashboards(request) is True
    assert not mock_grafana.dashboards


@responses.activate
def test_delete_nonexsiting_dashboard(mock_grafana):
    assert not mock_grafana.dashboards
    assert dashboard._delete_dashboard(mock_grafana.request, "2") is True


@responses.activate
def test_delete_dashboard_by_uid(mock_grafana):
    UID = "1"
    VERSION = 1
    TITLE = "testdashboard"
    dash = {"uid": UID, "title": TITLE, "version": VERSION}
    mock_grafana.create_dashboard(dash)
    request = mock_grafana.request

    assert mock_grafana.dashboards
    assert dashboard.delete_dashboard(request, dash)
    assert not mock_grafana.dashboards


@responses.activate
def test_delete_dashboard_by_title_and_folder(mock_grafana):
    FOLDER_ID = 1
    TITLE = "testdashboard"
    dash = {"title": TITLE, "folderId": FOLDER_ID}
    mock_grafana.create_dashboard(dash)
    request = mock_grafana.request

    assert mock_grafana.dashboards
    assert dashboard.delete_dashboard(request, dash)
    assert not mock_grafana.dashboards


@responses.activate
def test_search_dashboard(mock_grafana):
    UID = "1"
    TITLE = "testdashboard"
    mock_grafana.create_dashboard({"uid": UID, "title": TITLE})
    request = mock_grafana.request

    data = dashboard._search_dashboard(request, {"title": TITLE})
    assert data["uid"] == UID

    data = dashboard._search_dashboard(request, {'title': 'DoesNotExist'})
    assert data is None


@responses.activate
def test_search_dashboard_error(data_config):
    request = TokenRequest(**data_config, token='test')

    responses.add_callback(
        method=responses.GET,
        url=request.BASE_URL + 'api/search',
        callback=lambda f: (400, {}, ''))

    data = dashboard._search_dashboard(request, {'title': 'DoesNotExist'})
    assert data is None


@responses.activate
def test_create_dashboard(mock_grafana):
    UID = "1"
    VERSION = 1
    TITLE = "testdashboard"
    dashboard = {"uid": UID, "title": TITLE, "version": VERSION}
    request = mock_grafana.request
    assert not mock_grafana.dashboards

    provision.create_dashboard(request, dashboard)
    assert UID in mock_grafana.dashboards


@responses.activate
def test_create_dashboard_updates_with_existing_uid(mock_grafana):
    dashboard = {'title': 'testdashboard', 'version': 1}
    dash = mock_grafana.create_dashboard(dashboard)
    request = mock_grafana.request

    dashboard.update(**dash)
    dashboard['data'] = 'data'

    provision.create_dashboard(request, dashboard)
    in_grafana = mock_grafana.dashboards[dashboard['uid']]
    assert in_grafana['data'] == 'data'
    assert in_grafana['id'] == dashboard['id']


@responses.activate
def test_create_dashboard_no_uid_does_not_send_id(mock_grafana):
    ID = 1042
    VERSION = 1
    TITLE = 'testdashboard'
    dashboard = {'id': ID, 'title': TITLE, 'version': VERSION}
    request = mock_grafana.request

    data = provision.create_dashboard(request, dashboard)
    assert data['id'] != ID


@responses.activate
def test_create_dashboard_error(data_config):
    ID = 1042
    VERSION = 1
    TITLE = 'testdashboard'
    dashboard = {'id': ID, 'title': TITLE, 'version': VERSION}
    request = TokenRequest(**data_config, token='test')

    responses.add_callback(
        method=responses.GET,
        url=request.BASE_URL + 'api/search',
        callback=lambda f: (400, {}, ''))

    def post_callback(request):
        return 400, {}, ''

    responses.add_callback(
        method=responses.POST,
        url=request.BASE_URL + 'api/dashboards/db',
        callback=post_callback)

    data = provision.create_dashboard(request, dashboard)
    assert data is None
